/* start of file Laguer.c ----------------------------------------------- */

/* <summary>
	SUBROUTINE Laguer - Computes one root on the complex polynomial
	a = a[0], a[1],..., a[n-1]
   </summary>
*/

#include "Laguer.h"

float _Complex SetTo(float real, float imag);

void Laguer(float _Complex a[], int n, int *its, float _Complex *x)
/* x = a root of a, x input is the initial guess for a root
*/
/* The following routine implements the Laguerre method to find one
   root of a given polynomial of degree m = n-1, whose coefficients can
   be complex. The coefficient a[m] is the constant term, while a[0] is
   the coefficient of the highest power of x. The routine implements a
   simplified version of an elegant stopping criterion due to Adams
   [5], which neatly balances the desire to achieve full machine
   accuracy, on the one hand, with the danger of iterating forever in
   the presence of roundoff error, on the other.

   This routine improves x by Laguerre's method until it converges,
   within the achievable roundoff limit, to a root of the given
   polynomial. The number of iterations taken is returned as its.
   Parameters: EPSS is the estimated fractional roundoff error. We try
   to break any limit cycles with mR different fractional values, once
   every mT steps.
*/
{
	float EPSS = 2.0E-7F;	/* estimated fractional roundoff error */
	int m = n - 1;	/* n = size of polynomial, m = degree of polynomial */
	int mR = 8;
	int mT = 10;
	int iter, itToGo = 0, j;
	float abb;
	float abx;
	float abp;
	float abm;
	float _Complex cZero = SetTo(0.0F, 0.0F);
	float err;
	float _Complex mc = SetTo((float)m, 0.0F);
	float _Complex mmc = SetTo((float)(m - 1), 0.0F);
	float _Complex dx;
	float _Complex xx;
	float _Complex x1;
	float _Complex b;
	float _Complex d;
	float _Complex f;
	float _Complex g;
	float _Complex h;
	float _Complex sq;
	float _Complex gp;
	float _Complex gm;
	float _Complex g2;
	float frac[8];		/* Fractions used to break limit cycle. */
	frac[0] = 0.50F;
	frac[1] = 0.25F;
	frac[2] = 0.75F;
	frac[3] = 0.13F;
	frac[4] = 0.38F;
	frac[5] = 0.62F;
	frac[6] = 0.88F;
	frac[7] = 1.00F;
	xx = *x;				/* initial guess */
	float tf;
	/* Loop over iterations up to allowed max */
	for (iter = 1; ; iter++)
	{
		*its = iter;
		b = a[0];
		err = cabs(b);
		d = 0;
		f = 0;
		abx = cabs(xx);
		/* Efficient computation of the polynomial and its first */
		for (j = 1; j < n; j++)
		{
			f = xx * f + d;		/* two derivatives. f stores P''/2. */
			d = xx * d + b;
			b = xx * b + a[j];
			abb = cabs(b);
			err = abb + abx * err;
		}
		err *= EPSS;	/* Estimate of roundoff error in evaluating polynomial */
		if (abb <= err && itToGo == 0)
			itToGo = 2 * iter;
		if (b == cZero || (abb <= err && iter >= itToGo))		/* We are on the root */
		{
			int qq = 0;		/* for testing */
			qq++;
			break;
		}
		else	/* The generic case: use Laguerre's formula */
		{
			g = d / b;
			g2 = g * g;
			h = g2 - (f + f) / b;
			/* (m - 1) * (m * h - g2) */
			sq = mmc * (mc * h - g2);
			sq = csqrt(sq);
			gp = g + sq;
			gm = g - sq;
			abp = cabs(gp);
			abm = cabs(gm);
			if(abp < abm)
				gp = gm;
			tf = abp;
			if (abp < abm)	/* tf = Max(abp, abm) */
				tf = abm;
			if (tf > 0.0F)
				dx = mc / gp;
			else
			{
				tf = log(1.0F + abx);
				/* dx = tf + I * iter */
				dx = SetTo(tf, (float)iter);
				dx = cexp(dx);
			}
		}
		x1 = xx - dx;
		if (xx == x1)
			break;	/* Converged */
		j = iter % mT;
		if (j != 0)
			xx = x1;
		else	/* Every so often we take a fractional step, to break any */
				/* limit cycle (itself a rare occurrence) */
		{
			j = (-1 + iter / mT) % mR;
			xx -= dx * frac[j];
		}
	}
	*x = xx;
} /* Laguer */

/* end of file Laguer.c ----------------------------------------------- */


